import numpy as np

from open_rarity.models.collection import Collection, CollectionAttribute
from open_rarity.models.token import Token
from open_rarity.models.token_metadata import AttributeName
from open_rarity.scoring.utils import get_token_attributes_scores_and_weights


class HarmonicMeanScoringHandler:
    """harmonic mean of a token's n trait probabilities"""

    def __init__(self, normalized: bool = True):
        """
        Parameters
        ----------
        normalized : bool, optional
            If true, individual traits will be normalized based on total number
            of possible values for an attribute name, by default True.
        """
        self.normalized = normalized

    def score_token(self, collection: Collection, token: Token) -> float:
        return self._score_token(collection, token, self.normalized)

    def score_tokens(
        self,
        collection: Collection,
        tokens: list[Token],
    ) -> list[float]:
        # Memoize for performance
        collection_null_attributes = collection.extract_null_attributes()
        return [
            self._score_token(
                collection, t, self.normalized, collection_null_attributes
            )
            for t in tokens
        ]

    # Private methods
    def _score_token(
        self,
        collection: Collection,
        token: Token,
        normalized: bool = True,
        collection_null_attributes: dict[AttributeName, CollectionAttribute] = None,
    ) -> float:
        """Calculates the score of the token by taking the harmonic mean of the
        attribute scores with weights.

        Parameters
        ----------
        collection : Collection
            The collection with the attributes frequency counts to base the
            token trait probabilities on to calculate score.
        token : Token
            The token to score
        normalized : bool, optional
            Set to true to enable individual trait normalizations based on
            total number of possible values for an attribute name, by default True.
        collection_null_attributes : dict[AttributeName, CollectionAttribute], optional
            Optional memoization of collection.extract_null_attributes(),
            by default None.

        Returns
        -------
        float
            The token score
        """

        attr_scores, attr_weights = get_token_attributes_scores_and_weights(
            collection=collection,
            token=token,
            normalized=normalized,
            collection_null_attributes=collection_null_attributes,
        )

        return float(np.average(np.reciprocal(attr_scores), weights=attr_weights) ** -1)



import argparse
import csv
import io
import json
import logging
import math
import os
import pkgutil
from dataclasses import dataclass
from time import strftime, time
from typing import Iterable

import numpy as np

from open_rarity.models.collection import Collection
from open_rarity.models.token import Token
from open_rarity.models.token_identifier import EVMContractTokenIdentifier
from open_rarity.models.token_rarity import TokenRarity
from open_rarity.rarity_ranker import RarityRanker
from open_rarity.resolver.models.collection_with_metadata import CollectionWithMetadata
from open_rarity.resolver.models.token_with_rarity_data import (
    RankProvider,
    RarityData,
    TokenWithRarityData,
)
from open_rarity.resolver.opensea_api_helpers import (
    get_collection_with_metadata_from_opensea,
)
from open_rarity.resolver.rarity_providers.external_rarity_provider import (
    EXTERNAL_RANK_PROVIDERS,
    ExternalRarityProvider,
)
from open_rarity.scoring.handlers.arithmetic_mean_scoring_handler import (
    ArithmeticMeanScoringHandler,
)
from open_rarity.scoring.handlers.geometric_mean_scoring_handler import (
    GeometricMeanScoringHandler,
)
from open_rarity.scoring.handlers.harmonic_mean_scoring_handler import (
    HarmonicMeanScoringHandler,
)
from open_rarity.scoring.handlers.information_content_scoring_handler import (
    InformationContentScoringHandler,
)
from open_rarity.scoring.handlers.sum_scoring_handler import SumScoringHandler
from open_rarity.scoring.token_feature_extractor import TokenFeatureExtractor

harmonic_handler = HarmonicMeanScoringHandler()
arithmetic_handler = ArithmeticMeanScoringHandler()
geometric_handler = GeometricMeanScoringHandler()
sum_handler = SumScoringHandler()
ic_handler = InformationContentScoringHandler()

RankScore = tuple[int, float]
# Token ID -> Score
ScoredTokens = dict[int, float]
# Token ID -> Rank + Score
RankedTokens = dict[int, RankScore]

logger = logging.getLogger("open_rarity_logger")


parser = argparse.ArgumentParser()
parser.add_argument(
    "resolve_external_rarity",
    type=str,
    default="external",
    help="Specify 'external' if you want to resolve rarity from external providers",
)
parser.add_argument(
    "--cache",
    dest="cache_fetched_data",
    action=argparse.BooleanOptionalAction,
    default=True,
    help="Whether we use local data files to cache external trait + rank data",
)
parser.add_argument(
    "--filename",
    dest="filename",
    default="test_collections.json",
    help="File in /data folder containing collections to resolve.",
)

# The fastest external rarity provider, taking <15 seconds to resolve all rank data.
parser.add_argument(
    "--rarity_sniffer",
    dest="fetch_rarity_sniffer",
    action=argparse.BooleanOptionalAction,
    default=True,
    help="If external is specified, fetches rarity sniffer ranking data",
)

# Default disabled due to API key requirements. Without caching, this can take
# ~10 minutes for a 10k collection due to rate limits.
parser.add_argument(
    "--trait_sniper",
    dest="fetch_trait_sniper",
    action=argparse.BooleanOptionalAction,
    default=False,
    help=(
        "If external is specified, fetches trait sniper ranking data. "
        "If True, must set TRAIT_SNIPER_API_KEY env var."
    ),
)

# Default disabled because no public bulk fetcher function and therefore
# takes a long to fetch data for (~25 min for 10k collection).
parser.add_argument(
    "--rarity_sniper",
    dest="fetch_rarity_sniper",
    action=argparse.BooleanOptionalAction,
    default=False,
    help="If external is specified, fetches rarity sniper ranking data",
)


@dataclass
class OpenRarityScores:
    arithmetic_scores: RankedTokens
    geometric_scores: RankedTokens
    harmonic_scores: RankedTokens
    sum_scores: RankedTokens
    information_content_scores: RankedTokens


def get_tokens_with_rarity(
    collection_with_metadata: CollectionWithMetadata,
    external_rank_providers: list[RankProvider] = EXTERNAL_RANK_PROVIDERS,
    resolve_remote_rarity: bool = True,
    batch_size: int = 300,
    max_tokens_to_calculate: int = None,
    cache_external_ranks: bool = True,
) -> list[TokenWithRarityData]:
    """Resolves assets through OpenSea API asset endpoint and turns them
    into token with rarity data, augmented with rankings from Gem, RaritySniper
    and TraitSniper.

    Parameters
    ----------
    collection : Collection
        collection
    resolve_remote_rarity : bool
        True if we need to resolve rarity ranks from
        external providers , False if not
    max_tokens_to_calculate (int, optional): If specified only gets ranking
        data of first `max_tokens`. Defaults to None.
    cache_external_ranks : bool
        If set to true, will cache external ranks into local json file and
        optionally use cached data if file exists. If cache file already exists,
        will not refetch or rewrite cache data.

    Returns
    -------
    list[TokenWithRarityData]
        provide list of tokens augmented with assets metadata and ranking provider
    """
    slug = collection_with_metadata.opensea_slug
    external_rarity_provider = ExternalRarityProvider()
    total_supply = min(
        max_tokens_to_calculate or collection_with_metadata.token_total_supply,
        collection_with_metadata.token_total_supply,
    )
    num_batches = math.ceil(total_supply / batch_size)
    tokens = collection_with_metadata.collection.tokens
    if len(tokens) != collection_with_metadata.token_total_supply:
        # TODO [open-rarity] - Add support for collections with skipped token ids
        msg = (
            f"Warning: Collection {slug} has {len(tokens)} tokens, but total supply "
            f"is {collection_with_metadata.token_total_supply}. This can happen if "
            f"the collection skips token IDs. "
        )
        logger.warning(msg)
        print(msg)
    tokens_with_rarity: list[TokenWithRarityData] = []

    t1_start = time()

    tokens_batch: Iterable[Token]
    for batch_id, tokens_batch in enumerate(
        np.array_split(tokens, num_batches)  # type: ignore
    ):
        message = (
            f"\tStarting batch {batch_id} for collection "
            f"{slug}: Processing {len(tokens_batch)} tokens. "
            f"Last token: {tokens_batch[-1]}"
        )
        logger.debug(message)
        print(message)

        # We will store all rarities calculated across providers in this list
        tokens_rarity_batch = [
            TokenWithRarityData(token=t, rarities=[]) for t in tokens_batch
        ]

        if resolve_remote_rarity:
            external_rarity_provider.fetch_and_update_ranks(
                collection_with_metadata=collection_with_metadata,
                tokens_with_rarity=tokens_rarity_batch,
                rank_providers=external_rank_providers,
                cache_external_ranks=cache_external_ranks,
            )
        # Add the batch of augmented tokens with rarity into return value
        tokens_with_rarity.extend(tokens_rarity_batch)

    t1_stop = time()
    logger.debug(
        "Elapsed time during the asset resolution in seconds {seconds}".format(
            seconds=t1_stop - t1_start
        )
    )

    return tokens_with_rarity


def resolve_collection_data(
    resolve_remote_rarity: bool,
    external_rank_providers: list[RankProvider],
    package_path: str = "open_rarity.data",
    filename: str = "test_collections.json",
    max_tokens_to_calculate: int = None,
    use_cache: bool = True,
    output_file_to_disk: bool = True,
) -> list | None:
    """Resolves collection information through OpenSea API

    Parameters
    ----------
    resolve_remote_rarity : bool
        Set to true to resolve external rarity ranks for rank comparisons
    package_path : str, optional
        The package path for where the collection data to resolve collection data lives,
        by default "open_rarity.data"
    filename : str, optional
        The filename of the file holding the collection slugs to resolve,
        by default "test_collections.json"
    max_tokens_to_calculate : int, optional
        If specified only gets ranking data of first `max_tokens`, by default None.
        Note: If this is provided, we cannot calculate OpenRarity ranks since
        it must be calculated after calculating scoring for entire collection.
    use_cache: bool
        If set to true, will cache fetched data from external API's in order to ensure
        re-runs for same collections are faster. Only use if collection and token
        metadata is static - do not work for unrevealed/changing collections.
    output_file_to_disk: bool
        If set to true, will output the resolved collection data to disk.
        Set to False if you want to use the data in memory only.
        Needed for testing.

    Returns
    -------
    list
        A list of the rows that would be written to the output file. Only returned if
        output_file_to_disk is set to False.

    Raises
    ------
    ValueError
        If the file containing collection slugs to resolve does not exist.
    """
    golden_collections = pkgutil.get_data(package_path, filename)
    if not golden_collections:
        raise ValueError("Can't resolve golden collections data file.")

    data = json.load(io.BytesIO(golden_collections))
    print("------------------------------")
    output_rows: list = []
    for collection_def in data:
        start_time = time()
        opensea_slug = collection_def["collection_slug"]
        print(f"\nBEGIN: Resolving collection {opensea_slug}")
        print(
            f"1. Fetching collection & token traits for: {opensea_slug} from opensea."
        )
        # Fetch collection metadata and tokens that belong to this collection
        # from opensea and other external api's.
        collection_with_metadata = get_collection_with_metadata_from_opensea(
            opensea_collection_slug=opensea_slug,
            use_cache=use_cache,
        )
        print(f"2. Fetching external rarity ranks for: {opensea_slug}")
        tokens_with_rarity: list[TokenWithRarityData] = get_tokens_with_rarity(
            collection_with_metadata=collection_with_metadata,
            resolve_remote_rarity=resolve_remote_rarity,
            max_tokens_to_calculate=max_tokens_to_calculate,
            cache_external_ranks=use_cache,
            external_rank_providers=external_rank_providers,
        )
        print(f"3. Calculating OpenRarity ranks for: {opensea_slug}")

        collection = collection_with_metadata.collection

        if max_tokens_to_calculate is None:
            assert collection.token_total_supply == len(tokens_with_rarity)
        else:
            assert max_tokens_to_calculate == len(tokens_with_rarity)

        # Calculate and append open rarity scores
        if max_tokens_to_calculate is None:
            open_rarity_scores = resolve_open_rarity_score(
                collection, collection.tokens
            )
            augment_with_open_rarity_scores(
                tokens_with_rarity=tokens_with_rarity,
                scores=open_rarity_scores,
            )

        if output_file_to_disk:
            print(f"4. Wrote to CSV: {opensea_slug}")

        rows = serialize_to_csv(
            collection_with_metadata=collection_with_metadata,
            tokens_with_rarity=tokens_with_rarity,
            dry_run=not output_file_to_disk,
        )
        if rows is not None:
            output_rows += rows
        time_elapsed = round(time() - start_time)
        print(f"FINISHED: Resolved collection: {opensea_slug} in {time_elapsed} secs")

    return output_rows if not output_file_to_disk else None


def augment_with_open_rarity_scores(
    tokens_with_rarity: list[TokenWithRarityData], scores: OpenRarityScores
) -> None:
    """Augments tokens_with_rarity with ranks and scores computed by
    OpenRarity scorers'"""
    for token_with_rarity in tokens_with_rarity:
        token_id = token_with_rarity.token.token_identifier.token_id  # type: ignore

        token_with_rarity.rarities.extend(
            [
                RarityData(
                    provider=RankProvider.OR_ARITHMETIC,
                    rank=scores.arithmetic_scores[token_id][0],
                    score=scores.arithmetic_scores[token_id][1],
                ),
                RarityData(
                    provider=RankProvider.OR_GEOMETRIC,
                    rank=scores.geometric_scores[token_id][0],
                    score=scores.geometric_scores[token_id][1],
                ),
                RarityData(
                    provider=RankProvider.OR_HARMONIC,
                    rank=scores.harmonic_scores[token_id][0],
                    score=scores.harmonic_scores[token_id][1],
                ),
                RarityData(
                    provider=RankProvider.OR_SUM,
                    rank=scores.sum_scores[token_id][0],
                    score=scores.sum_scores[token_id][1],
                ),
                RarityData(
                    provider=RankProvider.OR_INFORMATION_CONTENT,
                    rank=scores.information_content_scores[token_id][0],
                    score=scores.information_content_scores[token_id][1],
                ),
            ]
        )


def extract_rank(tokens_to_score: dict[str, TokenRarity]) -> RankedTokens:
    """Sorts dictionary by float score and extract rank according to the score

    Parameters
    ----------
    token_id_to_scores : dict[str, TokenRarity]
        dictionary of token_id_to_scores with token_id to score mapping

    Returns
    -------
    dict[str, RankScore]
        dictionary of token to rank, score pair
    """
    ranked_tokens: list[TokenRarity] = RarityRanker.set_rarity_ranks(
        token_rarities=list(tokens_to_score.values())
    )

    result = {}
    for token in ranked_tokens:
        assert token.rank
        # note: this is a bug, ignoring the mypy error for now
        # token_identifier can be of type SolanaMintAddressTokenIdentifier
        # which has no token_id
        result[int(token.token.token_identifier.token_id)] = (  # type: ignore
            token.rank,
            token.score,
        )
    return result


def resolve_open_rarity_score(
    collection: Collection, tokens: list[Token]
) -> OpenRarityScores:
    """Resolve scores from all scorers with trait_normalization

    Parameters
    ----------
    collection : Collection
        collection is needed since the score is based on the invdividual's traits
        probability across the entire collection
    tokens: Subset of tokens belonging to Collection to resolve open rarity scores for

    """
    t1_start = time()

    # Dictionaries of token IDs to their respective TokenRarity for each strategy
    arthimetic_dict: dict[str, TokenRarity] = {}
    geometric_dict: dict[str, TokenRarity] = {}
    harmonic_dict: dict[str, TokenRarity] = {}
    sum_dict: dict[str, TokenRarity] = {}
    ic_dict: dict[str, TokenRarity] = {}

    logger.debug("OpenRarity scoring")

    for token in tokens:
        token_identifier = token.token_identifier
        assert isinstance(token_identifier, EVMContractTokenIdentifier)
        token_id = str(token_identifier.token_id)

        try:
            token_features = TokenFeatureExtractor.extract_unique_attribute_count(
                token=token, collection=collection
            )

            harmonic_dict[token_id] = TokenRarity(
                token=token,
                token_features=token_features,
                score=harmonic_handler.score_token(collection=collection, token=token),
            )
            arthimetic_dict[token_id] = TokenRarity(
                token=token,
                token_features=token_features,
                score=arithmetic_handler.score_token(
                    collection=collection, token=token
                ),
            )
            geometric_dict[token_id] = TokenRarity(
                token=token,
                token_features=token_features,
                score=geometric_handler.score_token(collection=collection, token=token),
            )
            sum_dict[token_id] = TokenRarity(
                token=token,
                token_features=token_features,
                score=sum_handler.score_token(collection=collection, token=token),
            )
            ic_dict[token_id] = TokenRarity(
                token=token,
                token_features=token_features,
                score=ic_handler.score_token(collection=collection, token=token),
            )

        except Exception:
            logger.exception(f"Can't score token {token} with OpenRarity")

    # Calculate ranks of all assets given the scores
    arthimetic_ranked_tokens = extract_rank(arthimetic_dict)
    geometric_ranked_tokens = extract_rank(geometric_dict)
    harmonic_ranked_tokens = extract_rank(harmonic_dict)
    sum_ranked_tokens = extract_rank(sum_dict)
    ic_ranked_tokens = extract_rank(ic_dict)

    t1_stop = time()
    logger.debug(
        "OpenRarity scores resolution in seconds {seconds}".format(
            seconds=t1_stop - t1_start
        )
    )

    return OpenRarityScores(
        arithmetic_scores=arthimetic_ranked_tokens,
        geometric_scores=geometric_ranked_tokens,
        harmonic_scores=harmonic_ranked_tokens,
        sum_scores=sum_ranked_tokens,
        information_content_scores=ic_ranked_tokens,
    )


def _get_provider_rank(
    provider: RankProvider, token_with_rarity: TokenWithRarityData
) -> int | None:
    """Get rank for the particular provider

    Parameters
    ----------
    provider : RankProvider
        rank provider
    token : Token
        token
    """
    rarities = token_with_rarity.rarities
    rarity_datas = list(filter(lambda rarity: rarity.provider == provider, rarities))
    return rarity_datas[0].rank if len(rarity_datas) > 0 else None


def _rank_diff(rank1: int | None, rank2: int | None) -> int | None:
    """Function that computes the rank difference

    Parameters
    ----------
    rank1 : int | None
        rank of the asset
    rank2 : int | None
        rank of the asset

    Returns
    -------
    int | None
        absolute difference of ranks for the specific asset
    """
    if not rank1 or not rank2:
        return None

    return abs(rank1 - rank2)


def serialize_to_csv(
    collection_with_metadata: CollectionWithMetadata,
    tokens_with_rarity: list[TokenWithRarityData],
    dry_run: bool = False,
) -> list | None:
    """Serialize collection and ranking data to CSV

    Parameters
    ----------
    collection_with_metadata : Collection
    dry_run: bool
        If set to True, the CSV will not be written to disk but returned as an
        array of rows

    """
    slug = collection_with_metadata.opensea_slug
    testset = open(f"testset_{slug}.csv", "w")
    headers = [
        "slug",
        "token_id",
        "traits_sniper",
        "rarity_sniffer",
        "rarity_sniper",
        "arithmetic",
        "geometric",
        "harmonic",
        "sum",
        "information_content",
        "traits_sniper_rarity_sniffer_diff",
        "traits_sniper_rarity_sniper_diff",
        "traits_sniper_arithm_diff",
        "traits_sniper_geom_diff",
        "traits_sniper_harmo_diff",
        "traits_sniper_sum_diff",
        "traits_sniper_ic_diff",
        "rarity_sniffer_rarity_sniper_diff",
        "rarity_sniffer_arithm_diff",
        "rarity_sniffer_geom_diff",
        "rarity_sniffer_harmo_diff",
        "rarity_sniffer_sum_diff",
        "rarity_sniffer_ic_diff",
        "rarity_sniper_arithm_diff",
        "rarity_sniper_geom_diff",
        "rarity_sniper_harmo_diff",
        "rarity_sniper_sum_diff",
        "rarity_sniper_ic_diff",
    ]

    writer = csv.writer(testset)
    writer.writerow(headers)
    rows = []

    for token_with_rarity in tokens_with_rarity:
        traits_sniper_rank = _get_provider_rank(
            RankProvider.TRAITS_SNIPER, token_with_rarity
        )
        rarity_sniffer_rank = _get_provider_rank(
            RankProvider.RARITY_SNIFFER, token_with_rarity
        )
        rarity_sniper_rank = _get_provider_rank(
            RankProvider.RARITY_SNIPER, token_with_rarity
        )
        or_arithmetic_rank = _get_provider_rank(
            RankProvider.OR_ARITHMETIC, token_with_rarity
        )
        or_geometric_rank = _get_provider_rank(
            RankProvider.OR_GEOMETRIC, token_with_rarity
        )
        or_harmonic_rank = _get_provider_rank(
            RankProvider.OR_HARMONIC, token_with_rarity
        )
        or_sum_rank = _get_provider_rank(RankProvider.OR_SUM, token_with_rarity)
        or_ic_rank = _get_provider_rank(
            RankProvider.OR_INFORMATION_CONTENT, token_with_rarity
        )
        row = [
            slug,
            token_with_rarity.token.token_identifier.token_id,  # type: ignore
            traits_sniper_rank,
            rarity_sniffer_rank,
            rarity_sniper_rank,
            or_arithmetic_rank,
            or_geometric_rank,
            or_harmonic_rank,
            or_sum_rank,
            or_ic_rank,
            _rank_diff(traits_sniper_rank, rarity_sniffer_rank),
            _rank_diff(traits_sniper_rank, rarity_sniper_rank),
            _rank_diff(traits_sniper_rank, or_arithmetic_rank),
            _rank_diff(traits_sniper_rank, or_geometric_rank),
            _rank_diff(traits_sniper_rank, or_harmonic_rank),
            _rank_diff(traits_sniper_rank, or_sum_rank),
            _rank_diff(traits_sniper_rank, or_ic_rank),
            _rank_diff(rarity_sniffer_rank, rarity_sniper_rank),
            _rank_diff(rarity_sniffer_rank, or_arithmetic_rank),
            _rank_diff(rarity_sniffer_rank, or_geometric_rank),
            _rank_diff(rarity_sniffer_rank, or_harmonic_rank),
            _rank_diff(rarity_sniffer_rank, or_sum_rank),
            _rank_diff(rarity_sniffer_rank, or_ic_rank),
            _rank_diff(rarity_sniper_rank, or_arithmetic_rank),
            _rank_diff(rarity_sniper_rank, or_geometric_rank),
            _rank_diff(rarity_sniper_rank, or_harmonic_rank),
            _rank_diff(rarity_sniper_rank, or_sum_rank),
            _rank_diff(rarity_sniper_rank, or_ic_rank),
        ]
        if dry_run:
            rows.append(row)
        else:
            writer.writerow(row)

    return rows if rows else None


if __name__ == "__main__":
    """Script to resolve external datasets and compute rarity scores
    on test collections. Data resolved from opensea api

    Command to run:
    `python -m open_rarity.resolver.testset_resolver external`

        This will only produce ranks from OpenRarity and RaritySniffer by default.

    To run for all external providers:
    `TRAIT_SNIPER_API_KEY=<your key> python -m open_rarity.resolver.testset_resolver \
        external --trait_sniper --rarity_sniper`
    """
    args = parser.parse_args()
    logger = logging.getLogger("open_rarity_logger")
    logger.setLevel(logging.DEBUG)

    fh = logging.FileHandler(strftime("testsetresolverlog_%H_%M_%m_%d_%Y.log"))
    fh.setLevel(logging.DEBUG)

    logger.addHandler(fh)
    resolve_remote_rarity = args.resolve_external_rarity == "external"

    external_resolvers = []
    if args.fetch_trait_sniper:
        external_resolvers.append(RankProvider.TRAITS_SNIPER)
        if os.environ.get("TRAIT_SNIPER_API_KEY") is None:
            raise ValueError(
                "TRAIT_SNIPER_API_KEY not set. Required to fetch Traits Sniper data"
            )
    if args.fetch_rarity_sniffer:
        external_resolvers.append(RankProvider.RARITY_SNIFFER)
    if args.fetch_rarity_sniper:
        external_resolvers.append(RankProvider.RARITY_SNIPER)

    print(
        "Welcome to OpenRarity resolver! This is a tool to view OpenRarity rankings \n"
        "for given collection(s) and to compare them with existing ranking \n"
        "providers. If you just want to output OpenRarity rankings, you can use "
        "the script scripts/score_real_collections.py. \nFor full options, "
        "run `python3 -m open_rarity.resolver.testset_resolver --help`"
    )
    print(f"\nExecuting args: {args}")
    if resolve_remote_rarity:
        print(
            f"Fetching external ranks from: {[rp.value for rp in external_resolvers]}"
        )

    print(
        "\nNOTE: Resolving external data can take awhile due to external API rate"
        "\nlimits. Local caching will occur automatically so that future runs of "
        "\nthe same collection can be efficient. Expect a 10k collection to take >5 min"
        "\nwithout local cached data (timing based on exact external resolver(s) set). "
        "\nWith caching, expect ~15 seconds for processing."
    )

    resolve_collection_data(
        resolve_remote_rarity,
        external_rank_providers=external_resolvers,
        use_cache=args.cache_fetched_data,
        filename=args.filename,
    )



from dataclasses import dataclass


@dataclass
class TokenRankingFeatures:
    """Class represents all standardized ranking features
    that should be considered by the ranking function.

    Attributes
    ----------
    unique_attribute_count : int
        count of unique attributes in the token
    """

    unique_attribute_count: int



import logging

import requests

from .rank_resolver import RankResolver

logger = logging.getLogger("open_rarity_logger")
RARITY_SNIPER_API_URL = (
    "https://api.raritysniper.com/public/collection/{slug}/id/{token_id}"
)
USER_AGENT = {
    "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/96.0.4664.45 Safari/537.36"  # noqa: E501
}


class RaritySniperResolver(RankResolver):
    @staticmethod
    def get_all_ranks(slug: str) -> dict[str, int]:
        raise NotImplementedError

    @staticmethod
    def get_slug(opensea_slug: str) -> str:
        # custom fixes to normalize slug name
        # used in rarity sniper
        slug = opensea_slug.replace("-nft", "")
        slug = slug.replace("-official", "")
        slug = slug.replace("beanzofficial", "beanz")
        slug = slug.replace("boredapeyachtclub", "bored-ape-yacht-club")
        slug = slug.replace("clonex", "clone-x")
        slug = slug.replace("invisiblefriends", "invisible-friends")
        slug = slug.replace("proof-", "")
        slug = slug.replace("pudgypenguins", "pudgy-penguins")
        slug = slug.replace("wtf", "")

        return slug

    @staticmethod
    def get_rank(collection_slug: str, token_id: int) -> int | None:
        url = RARITY_SNIPER_API_URL.format(slug=collection_slug, token_id=token_id)
        logger.debug("{url}".format(url=url))
        response = requests.request("GET", url, headers=USER_AGENT)
        if response.status_code == 200:
            return response.json()["rank"]
        else:
            logger.debug(
                f"[RaritySniper] Failed to resolve Rarity Sniper rank for "
                f"{collection_slug} {token_id}. Received {response.status_code} for "
                f"{url}: {response.reason}. {response.json()}"
            )
            return None



from open_rarity.models.collection import Collection, CollectionAttribute
from open_rarity.models.token import Token
from open_rarity.models.token_metadata import AttributeName


def get_token_attributes_scores_and_weights(
    collection: Collection,
    token: Token,
    normalized: bool,
    collection_null_attributes: dict[AttributeName, CollectionAttribute] = None,
) -> tuple[list[float], list[float]]:
    """Calculates the scores and normalization weights for a token
    based on its attributes. If the token does not have an attribute, the probability
    of the attribute being null is used instead.

    Parameters
    ----------
    collection : Collection
        The collection to calculate probability on.
    token : Token
        The token to score.
    normalized : bool
        Set to true to enable individual trait normalizations based on total
        number of possible values for an attribute, by default True.
    collection_null_attributes : dict[ AttributeName, CollectionAttribute ], optional
        Optional memoization of collection.extract_null_attributes(), by default None.

    Returns
    -------
    tuple[list[float], list[float]]
        A tuple of attribute scores and attribute weights.
        attribute scores: scores for an attribute is defined to be the inverse of
            the probability of that attribute existing across the collection. e.g.
            (total token supply / total tokens with that attribute name and value)
        attribute weights: The weights for each score that should be applied
            if normalization is to occur.
    """
    # Create a combined attributes dictionary such that if the token has the attribute,
    # it uses the value's probability, and if it doesn't have the attribute,
    # uses the probability of that attribute being null.
    if collection_null_attributes is None:
        null_attributes = collection.extract_null_attributes()
    else:
        null_attributes = collection_null_attributes

    combined_attributes: dict[
        str, CollectionAttribute
    ] = null_attributes | _convert_to_collection_attributes_dict(collection, token)

    sorted_attr_names = sorted(list(combined_attributes.keys()))
    sorted_attrs = [combined_attributes[attr_name] for attr_name in sorted_attr_names]

    total_supply = collection.token_total_supply

    # Normalize traits by dividing by the total number of possible values for
    # that trait. The normalization factor takes into account the cardinality
    # values for particual traits, such that high cardinality traits aren't
    # over-indexed in rarity.
    # Example: If Asset has a trait "Hat" and it has possible values
    # {"Red","Yellow","Green"} the normalization factor will be 1/3 or
    # 0.33. If a trait has 10,000 options, than the normalization factor is 1/10,000.
    if normalized:
        attr_weights = [
            1 / collection.total_attribute_values(attr_name)
            for attr_name in sorted_attr_names
        ]
    else:
        attr_weights = [1.0] * len(sorted_attr_names)

    scores = [total_supply / attr.total_tokens for attr in sorted_attrs]

    return (scores, attr_weights)


def _convert_to_collection_attributes_dict(collection, token):
    # NOTE: We currently only support string attributes
    return {
        attribute.name: CollectionAttribute(
            attribute=attribute,
            total_tokens=collection.total_tokens_with_attribute(attribute),
        )
        for attribute in token.metadata.string_attributes.values()
    }



from dataclasses import dataclass
from typing import Annotated, Literal, Type, TypeAlias, Union

from pydantic import Field


@dataclass(frozen=True)
class EVMContractTokenIdentifier:
    """This token is identified by the contract address and token ID number.

    This identifier is based off of the interface as defined by ERC721 and ERC1155,
    where unique tokens belong to the same contract but have their own numeral token id.
    """

    contract_address: str
    token_id: int
    identifier_type: Literal["evm_contract"] = "evm_contract"

    def __str__(self):
        return f"Contract({self.contract_address}) #{self.token_id}"

    @classmethod
    def from_dict(cls, data_dict: dict):
        return cls(
            contract_address=data_dict["contract_address"],
            token_id=data_dict["token_id"],
        )

    def to_dict(self) -> dict:
        return {
            "contract_address": self.contract_address,
            "token_id": self.token_id,
        }


@dataclass(frozen=True)
class SolanaMintAddressTokenIdentifier:
    """This token is identified by their solana account address.

    This identifier is based off of the interface defined by the Solana SPL token
    standard where every such token is declared by creating a mint account.
    """

    mint_address: str
    identifier_type: Literal["solana_mint_address"] = "solana_mint_address"

    def __str__(self):
        return f"MintAddress({self.mint_address})"

    @classmethod
    def from_dict(cls, data_dict: dict):
        return cls(
            mint_address=data_dict["mint_address"],
        )

    def to_dict(self) -> dict:
        return {
            "mint_address": self.mint_address,
        }


# This is used to specifies how the collection is identified and the
# logic used to group the NFTs together
TokenIdentifier = Annotated[
    (EVMContractTokenIdentifier | SolanaMintAddressTokenIdentifier),
    Field(discriminator="identifier_type"),
]

TokenIdentifierClass: TypeAlias = Union[
    Type[EVMContractTokenIdentifier], Type[SolanaMintAddressTokenIdentifier]
]


def get_identifier_class_from_dict(data_dict: dict) -> TokenIdentifierClass:
    return (
        EVMContractTokenIdentifier
        if "token_id" in data_dict
        else SolanaMintAddressTokenIdentifier
    )



import json
import logging
from collections import defaultdict

from open_rarity.models.token_identifier import EVMContractTokenIdentifier
from open_rarity.resolver.models.collection_with_metadata import CollectionWithMetadata
from open_rarity.resolver.models.token_with_rarity_data import (
    EXTERNAL_RANK_PROVIDERS,
    RankProvider,
    RarityData,
    TokenWithRarityData,
)

from .rank_resolver import RankResolver
from .rarity_sniffer import RaritySnifferResolver
from .rarity_sniper import RaritySniperResolver
from .trait_sniper import TraitSniperResolver

logger = logging.getLogger("open_rarity_logger")


def get_external_resolver(rank_provider: RankProvider) -> RankResolver:
    if rank_provider == RankProvider.TRAITS_SNIPER:
        return TraitSniperResolver()
    elif rank_provider == RankProvider.RARITY_SNIFFER:
        return RaritySnifferResolver()
    elif rank_provider == RankProvider.RARITY_SNIPER:
        return RaritySniperResolver()
    raise Exception(f"Unknown external rank provider: {rank_provider}")


class ExternalRarityProvider:
    # Cached file will have format:
    # "{collection_slug}_{provider_name}_cached_ranks.json"
    # The data must be a dictionary of <token id as int> to <rank as int>
    CACHE_FILENAME_FORMAT: str = "cached_data/%s_%s_cached_ranks.json"

    # Dictionary of slug -> {token_id (str) -> rank (int)}
    _trait_sniper_cache: dict[str, dict[str, int]] = defaultdict(dict)
    _rarity_sniffer_cache: dict[str, dict[str, int]] = defaultdict(dict)
    _rarity_sniper_cache: dict[str, dict[str, int]] = defaultdict(dict)

    def cache_filename(self, rank_provider: RankProvider, slug: str) -> str:
        rank_name = rank_provider.name.lower()
        return self.CACHE_FILENAME_FORMAT % (slug, rank_name)

    def fetch_and_update_ranks(
        self,
        collection_with_metadata: CollectionWithMetadata,
        tokens_with_rarity: list[TokenWithRarityData],
        rank_providers: list[RankProvider] = EXTERNAL_RANK_PROVIDERS,
        cache_external_ranks: bool = True,
    ) -> list[TokenWithRarityData]:
        """Fetches ranks from available providers gem, rarity sniper and/or trait sniper
        and adds them to the rarities field in `tokens_with_rarity`

        Parameters
        ----------
        collection : Collection
            collection
        tokens_with_rarity: list[TokenWithRaritydata]
            List of tokens with rarity data. Will modify the objects.rarities
            field and add the fetched ranking data directly to object.
        cache_external_ranks: bool
            If set to true, will use local cache file instead of fetching rank data.
            If cache is empty, will fetch data from API and write to local cache.

        Returns
        -------
        list[tokens_with_rarity]
            tokens with fetched external rarity data
        """
        logger.debug(f"Fetching external rarity for {len(tokens_with_rarity)} tokens")

        for rank_provider in rank_providers:
            # Note: Not all providers have rankings for all collections,
            # so do a best effort.
            print("\t\tProcessing provider: ", rank_provider)
            try:
                if rank_provider in {
                    RankProvider.RARITY_SNIFFER,
                    RankProvider.TRAITS_SNIPER,
                }:
                    self._add_rarity_data(
                        rank_provider=rank_provider,
                        collection_with_metadata=collection_with_metadata,
                        tokens_with_rarity=tokens_with_rarity,
                        cache_external_ranks=cache_external_ranks,
                    )
                elif rank_provider == RankProvider.RARITY_SNIPER:
                    self._add_rarity_sniper_rarity_data(
                        collection_with_metadata=collection_with_metadata,
                        tokens_with_rarity=tokens_with_rarity,
                        cache_external_ranks=cache_external_ranks,
                    )
            except Exception:
                logger.exception(
                    f"Exception: Could not get ranks from {rank_provider} for "
                    f"{collection_with_metadata.opensea_slug}",
                    exc_info=True,
                )
                continue
        return tokens_with_rarity

    # Private methods
    def _add_rarity_data(
        self,
        rank_provider: RankProvider,
        collection_with_metadata: CollectionWithMetadata,
        tokens_with_rarity: list[TokenWithRarityData],
        cache_external_ranks: bool = True,
    ) -> list[TokenWithRarityData]:
        """Modifies `tokens_with_rarity` by adding external provider rank data
        Currently only works for EVM collections

        Parameters
        ----------
        collection_with_metadata : CollectionWithMetadata
        tokens_with_rarity : list[TokenWithRarityData]
            list of tokens with rarity to augment

        Returns
        -------
        list[TokenWithRarityData]
            list of augmeneted tokens
        """
        slug = collection_with_metadata.opensea_slug
        logger.debug(f"Resolving {rank_provider} data for {slug}")
        contract_addresses = collection_with_metadata.contract_addresses
        if len(contract_addresses) != 1:
            raise ValueError(
                f"We cannot calculate {rank_provider.value} score for collections "
                f"that do not map to a single contract address: {contract_addresses=}"
            )
        contract_address = contract_addresses[0]

        # Get input params and validate
        if rank_provider == RankProvider.RARITY_SNIPER:
            raise ValueError(
                "Invalid rank provider. Needs to support bulk rank fetching first."
            )

        # Load data from local cache if available
        if cache_external_ranks:
            self._load_cache_from_file(slug=slug, rank_provider=rank_provider)

        # If we didn't want to load cache or cache is empty, pull data from API
        if not self._is_cache_loaded(slug, rank_provider):
            resolver = get_external_resolver(rank_provider)
            token_ids_to_ranks = resolver.get_all_ranks(contract_address)
            self._set_cache(
                slug=slug, rank_provider=rank_provider, rank_data=token_ids_to_ranks
            )
            num_tokens = len(token_ids_to_ranks)
            logger.debug(
                f"Fetched {num_tokens} ranks from {rank_provider} API and cached"
            )
            if not token_ids_to_ranks:
                logger.warning(f"Did not get any data for {slug} from {rank_provider}")
                raise Exception(
                    f"Could not get data from external rank provider: {rank_provider}"
                )
            if cache_external_ranks:
                self.write_cache_to_file(slug, rank_provider)

        # Fill in ranks for each token
        for token_with_rarity in tokens_with_rarity:
            token_identifer = token_with_rarity.token.token_identifier
            # Needed for type-checking
            assert isinstance(token_identifer, EVMContractTokenIdentifier)
            token_id = token_identifer.token_id

            rank = self._get_cached_rank(
                slug=slug, rank_provider=rank_provider, token_id=token_id
            )

            if rank:
                token_with_rarity.rarities.append(
                    RarityData(provider=rank_provider, rank=rank)
                )

        return tokens_with_rarity

    def _add_rarity_sniper_rarity_data(
        self,
        collection_with_metadata: CollectionWithMetadata,
        tokens_with_rarity: list[TokenWithRarityData],
        cache_external_ranks: bool = True,
    ) -> list[TokenWithRarityData]:
        # We're currently using opensea slug to calculate trait sniper slug
        opensea_slug = collection_with_metadata.opensea_slug
        slug = RaritySniperResolver.get_slug(opensea_slug=opensea_slug)
        rank_provider = RankProvider.RARITY_SNIPER
        if cache_external_ranks:
            self._load_cache_from_file(slug=opensea_slug, rank_provider=rank_provider)

        for token_with_rarity in tokens_with_rarity:
            token = token_with_rarity.token
            token_identifer = token.token_identifier
            # Needed for type-checking
            assert isinstance(token_identifer, EVMContractTokenIdentifier)
            token_id = token_identifer.token_id
            rank = self._get_cached_rank(
                slug=opensea_slug,
                rank_provider=rank_provider,
                token_id=token_id,
            )

            if rank is None:
                try:
                    rank = RaritySniperResolver.get_rank(
                        collection_slug=slug, token_id=token_id
                    )
                    print(
                        f"[RaritySniper] Fetched from api: {rank=} {token_id=} {slug=}"
                    )
                    logger.debug(
                        "Resolved rarity sniper rarity for "
                        f"{opensea_slug=}/{slug=} {token_id=}: {rank}"
                    )
                except Exception:
                    logger.exception(
                        "[Rarity Sniper] Failed to resolve from API:"
                        f"{opensea_slug=}/{slug=} {token_id=}: {rank}",
                        exc_info=True,
                    )

                # Store in cache
                if rank is not None:
                    self._get_cache_for_collection(opensea_slug, rank_provider)[
                        str(token_id)
                    ] = rank

            if rank:
                token_with_rarity.rarities.append(
                    RarityData(provider=rank_provider, rank=rank)
                )

        if cache_external_ranks:
            self.write_cache_to_file(opensea_slug, rank_provider)

        return tokens_with_rarity

    # Cache methods
    def _load_cache_from_file(
        self,
        slug: str,
        rank_provider: RankProvider,
        force_reload: bool = False,
    ) -> bool:
        # Short-circuit if cache is already loaded, unless we want to force a reload
        if not force_reload and self._is_cache_loaded(slug, rank_provider):
            return False

        cache_filename = self.cache_filename(rank_provider=rank_provider, slug=slug)
        try:
            with open(cache_filename) as jsonfile:
                external_rank_data = json.load(jsonfile)
            logger.debug(
                f"Successfully loaded cached external ranks from: {cache_filename}: "
                f"Found {len(external_rank_data)} token ranks"
            )
        except FileNotFoundError:
            logger.warning(f"Cache file does not exist: {cache_filename}.")
            return False
        except Exception:
            logger.exception(
                f"Could not parse cache file: {cache_filename}.", exc_info=True
            )
            return False

        self._set_cache(
            slug=slug, rank_provider=rank_provider, rank_data=external_rank_data
        )
        return True

    def write_cache_to_file(self, slug: str, rank_provider: RankProvider):
        cache_data = self._get_cache_for_collection(
            slug=slug, rank_provider=rank_provider
        )
        cache_filename = self.cache_filename(rank_provider=rank_provider, slug=slug)
        logger.debug(
            f"Writing external rank data ({rank_provider}) to cache for: {slug} "
            f"to file: {cache_filename}. Contains {len(cache_data)} token ranks."
        )
        with open(cache_filename, "w+") as jsonfile:
            json.dump(cache_data, jsonfile, indent=4)

    def _set_cache(
        self, slug: str, rank_provider: RankProvider, rank_data: dict[str, int]
    ) -> None:
        self._get_cache(rank_provider)[slug] = rank_data

    def _get_cache(self, rank_provider: RankProvider) -> dict[str, dict[str, int]]:
        if rank_provider == RankProvider.TRAITS_SNIPER:
            return self._trait_sniper_cache
        if rank_provider == RankProvider.RARITY_SNIFFER:
            return self._rarity_sniffer_cache
        if rank_provider == RankProvider.RARITY_SNIPER:
            return self._rarity_sniper_cache
        raise Exception(f"Unknown external rank provider: {rank_provider}")

    def _get_cache_for_collection(
        self, slug: str, rank_provider: RankProvider
    ) -> dict[str, int]:
        return self._get_cache(rank_provider)[slug]

    def _get_cached_rank(
        self, slug: str, rank_provider: RankProvider, token_id: int
    ) -> int | None:
        return self._get_cache_for_collection(slug, rank_provider).get(
            str(token_id), None
        )

    def _is_cache_loaded(self, slug: str, rank_provider: RankProvider):
        return (
            len(self._get_cache_for_collection(rank_provider=rank_provider, slug=slug))
            > 0
        )



import json
import logging
import math
import os

import requests
from requests.models import HTTPError

from open_rarity.models.collection import Collection
from open_rarity.models.token import Token
from open_rarity.models.token_identifier import EVMContractTokenIdentifier
from open_rarity.models.token_metadata import (
    DateAttribute,
    NumericAttribute,
    StringAttribute,
    TokenMetadata,
)
from open_rarity.models.token_standard import TokenStandard
from open_rarity.resolver.models.collection_with_metadata import CollectionWithMetadata

logger = logging.getLogger("open_rarity_logger")

# https://docs.opensea.io/reference/retrieving-a-single-collection
OS_COLLECTION_URL = "https://api.opensea.io/api/v1/collection/{slug}"
OS_ASSETS_URL = "https://api.opensea.io/api/v1/assets"

HEADERS = {
    "Accept": "application/json",
    # Note: API key not required but will help rate limiting
    "X-API-KEY": os.environ.get("OS_API_KEY") or "",
}

# https://docs.opensea.io/docs/metadata-standards
OS_METADATA_TRAIT_TYPE = "display_type"

# Reads from a local cache of format:
# "cached_data/{slug}_cached_os_trait_data.json"
# which must contain a list of tokens in Token.to_dict() format.
# See cached_data/boredapeyachtclub_cached_os_trait_data.json for example.
OS_CACHE_FILENAME_FORMAT: str = "cached_data/%s_cached_os_trait_data.json"


# Error is thrown if computatation is requested on a non-ERC721/1155
# token or collection. This is due to library only working for these
# standards currently.
class ERCStandardError(ValueError):
    pass


def fetch_opensea_collection_data(slug: str) -> dict:
    """Fetches collection data from Opensea's GET collection endpoint for
    the given slug.

    Raises:
        Exception: If API request fails
    """
    response = requests.get(OS_COLLECTION_URL.format(slug=slug), headers=HEADERS)

    if response.status_code != 200:
        logger.debug(
            f"[Opensea] Failed to resolve collection {slug}."
            f"Received {response.status_code}: {response.reason}. {response.json()}"
        )

        response.raise_for_status()

    return response.json()["collection"]


def fetch_opensea_assets_data(slug: str, token_ids: list[int], limit=30) -> list[dict]:
    """Fetches asset data from Opensea's GET assets endpoint for the given token ids

    Parameters
    ----------
    slug: str
        Opensea collection slug
    token_ids: list[int]
        the token id
    limit: int, optional
        How many to fetch at once. Defaults to 30, with a max of 30, by default 30.

    Returns
    -------
    list[dict]
        list of asset data dictionaries, e.g. the response in "assets" field,
        sorted by token_id asc

    Raises
    ------
        Exception: If api request fails


    """
    assert len(token_ids) <= limit
    # Max 30 limit enforced on API
    assert limit <= 30
    querystring = {
        "token_ids": token_ids,
        "collection_slug": slug,
        "offset": "0",
        "limit": limit,
    }

    response = requests.request(
        "GET",
        OS_ASSETS_URL,
        headers=HEADERS,
        params=querystring,
    )

    if response.status_code != 200:
        logger.debug(
            f"[Opensea] Failed to resolve assets for {slug}."
            f"Received {response.status_code}: {response.reason}. {response.json()}"
        )
        response.raise_for_status()

    # The API does not sort return value assets by token ID, so sort then return
    return sorted(response.json()["assets"], key=(lambda a: int(a["token_id"])))


def opensea_traits_to_token_metadata(asset_traits: list) -> TokenMetadata:
    """Converts asset traits list returned by opensea assets API and converts
    it into a TokenMetadata.

    Parameters
    ----------
    asset_traits : list
        the "traits" field for an asset in the return value of Opensea's asset(s)
        endpoint.

    Returns
    -------
    TokenMetadata
        A TokenMetadata instance to hold the token metadata extracted from the
        input data.
    """

    string_attr = {}
    numeric_attr = {}
    date_attr = {}

    for trait in asset_traits:
        if is_string_trait(trait):
            string_attr[trait["trait_type"]] = StringAttribute(
                name=trait["trait_type"],
                value=trait["value"],
            )
        elif is_numeric_trait(trait):
            numeric_attr[trait["trait_type"]] = NumericAttribute(
                name=trait["trait_type"],
                value=trait["value"],
            )
        elif is_date_trait(trait):
            date_attr[trait["trait_type"]] = DateAttribute(
                name=trait["trait_type"],
                value=trait["value"],
            )
        else:
            logger.debug(f"Unknown trait type {trait}")

    return TokenMetadata(
        string_attributes=string_attr,
        numeric_attributes=numeric_attr,
        date_attributes=date_attr,
    )


def get_all_collection_tokens(
    slug: str,
    total_supply: int,
    batch_size: int = 30,
    use_cache: bool = True,
) -> list[Token]:
    """Returns a list of Token's with all metadata filled, either populated
    from Opensea API or fetched from a local cache file.
    """
    tokens: list[Token] = []

    # For performance optimization and re-runs for the same collection,
    # we optionally check if an already cached file for the collection
    # data exists since they tend to be static.
    if use_cache:
        tokens = read_collection_data_from_file(expected_supply=total_supply, slug=slug)
    else:
        logger.info(f"Not using cache for fetching collection tokens for: {slug}")

    # This means either cache file didn't exist or did not have data.
    # Fetch all token trait data from opensea.
    if len(tokens) == 0:
        tokens = []
        num_batches = math.ceil(total_supply / batch_size)
        initial_token_id = 0

        # Returns a list of `batch_size` token IDs, such that no token ID
        # can exceed `max_token_id` (in which case len(return_value) < `batch_size`)
        def get_token_ids(
            batch_id: int, max_token_id: int = total_supply - 1
        ) -> list[int]:
            token_id_start = initial_token_id + (batch_id * batch_size)
            token_id_end = int(min(token_id_start + batch_size - 1, max_token_id))
            return list(range(token_id_start, token_id_end + 1))

        for batch_id in range(num_batches):
            token_ids = get_token_ids(batch_id)
            tokens_batch = get_tokens_from_opensea(
                opensea_slug=slug,
                token_ids=token_ids,
            )

            tokens.extend(tokens_batch)

        # It's possible for some collections to start at token id 1 instead of 0,
        # so attempt fetch of more tokens if they exist
        token_id = total_supply
        while True:
            try:
                extra_tokens = get_tokens_from_opensea(
                    opensea_slug=slug,
                    token_ids=[token_id],
                )
                if len(extra_tokens) == 0:
                    break
                tokens.extend(extra_tokens)
                token_id += 1
            except Exception:
                break

        if len(tokens) > total_supply:
            logger.warning(
                f"Warning: Found more tokens ({len(tokens)}) than "
                f"token supply ({total_supply}) fetched from collection stats"
            )

        # Write to local disk the fetched data for later caching
        if use_cache:
            write_collection_data_to_file(slug=slug, tokens=tokens)

    return tokens


def get_tokens_from_opensea(opensea_slug: str, token_ids: list[int]) -> list[Token]:
    """Fetches eth nft data from opensea API and stores them into Token objects

    Parameters
    ----------
    opensea_slug : str
        Opensea collection slug
    token_ids : list[int]
        List of token ids to fetch for

    Returns
    -------
    list[Token]
        Returns list of tokens if request is successful.

    Raises
    ------
    ValueError
        if request is made to a non ERC721 or ERC1155 collection
    HTTPError
        if request to opensea fails
    """
    try:
        assets = fetch_opensea_assets_data(slug=opensea_slug, token_ids=token_ids)
    except HTTPError as e:
        logger.exception(
            "FAILED: get_assets: could not fetch opensea assets for %s: %s",
            token_ids,
            e,
            exc_info=True,
        )
        raise

    tokens = []
    for asset in assets:
        token_metadata = opensea_traits_to_token_metadata(asset_traits=asset["traits"])
        asset_contract_address = asset["asset_contract"]["address"]
        asset_contract_type = asset["asset_contract"]["asset_contract_type"]
        if asset_contract_type == "non-fungible":
            token_standard = TokenStandard.ERC721
        elif asset_contract_type == "semi-fungible":
            token_standard = TokenStandard.ERC1155
        else:
            raise ValueError(f"Unexpected asset contrat type: {asset_contract_type}")
        tokens.append(
            Token(
                token_identifier=EVMContractTokenIdentifier(
                    identifier_type="evm_contract",
                    contract_address=asset_contract_address,
                    token_id=int(asset["token_id"]),
                ),
                token_standard=token_standard,
                metadata=token_metadata,
            )
        )

    return tokens


def get_collection_with_metadata_from_opensea(
    opensea_collection_slug: str,
    use_cache: bool,
) -> CollectionWithMetadata:
    """Fetches collection metadata with OpenSea endpoint and API key
    and stores it in the Collection object with 0 tokens.

    Parameters
    ----------
    opensea_collection_slug : str
        collection slug on opensea's system
    use_cache: bool
        If true, reads the token trait data from local cache file if it exists, or
        fetches from opensea api and stores the data in a local cached file for
        future reuse.

    Returns
    -------
    CollectionWithMetadata
        collection with metadata, but with no tokens

    """
    collection_obj = fetch_opensea_collection_data(slug=opensea_collection_slug)
    contracts = collection_obj["primary_asset_contracts"]
    interfaces = set([contract["schema_name"] for contract in contracts])
    stats = collection_obj["stats"]
    if not interfaces.issubset(set(["ERC721", "ERC1155"])):
        raise ERCStandardError(
            "We currently do not support non EVM standards at the moment"
        )
    total_supply = int(stats["total_supply"])
    tokens = get_all_collection_tokens(
        slug=opensea_collection_slug,
        total_supply=total_supply,
        use_cache=use_cache,
    )

    collection = Collection(
        name=collection_obj["name"],
        attributes_frequency_counts=collection_obj["traits"],
        tokens=tokens,
    )

    collection_with_metadata = CollectionWithMetadata(
        collection=collection,
        contract_addresses=[contract["address"] for contract in contracts],
        token_total_supply=max(total_supply, collection.token_total_supply),
        opensea_slug=opensea_collection_slug,
    )

    return collection_with_metadata


def get_collection_from_opensea(
    slug: str, batch_size: int = 30, use_cache: bool = True
) -> Collection:
    """Fetches collection and token data with OpenSea endpoint and API key
    and stores it in the Collection object. If local cache file is used and
    contains all collection token data, that data will be used instead.

    Parameters
    ----------
    slug : str
        collection slug on opensea's system

    batch_size: int
        batch size for the opensea API requests
        maximum value is 30

    use_cache: bool
        set to True to look for a local cached version of the collection and token
        metadata fetched from opensea to prevent re-fetching.

    Returns
    -------
    Collection
        collection abstraction

    """
    # Fetch collection metadata
    collection_obj = fetch_opensea_collection_data(slug=slug)
    contracts = collection_obj["primary_asset_contracts"]
    interfaces = set([contract["schema_name"] for contract in contracts])
    stats = collection_obj["stats"]
    if not interfaces.issubset(set(["ERC721"])):
        raise ERCStandardError(
            "We currently do not support non ERC721 standards at the moment"
        )

    total_supply = int(stats["total_supply"])
    tokens = get_all_collection_tokens(
        slug=slug,
        total_supply=total_supply,
        batch_size=batch_size,
        use_cache=use_cache,
    )

    return Collection(name=collection_obj["name"], tokens=tokens)


def get_cache_filename(slug: str) -> str:
    return OS_CACHE_FILENAME_FORMAT % slug


def write_collection_data_to_file(slug: str, tokens: list[Token]):
    cache_filename = get_cache_filename(slug)
    json_output = []
    for token in tokens:
        # Note: We assume EVM token here
        json_output.append(token.to_dict())
    with open(cache_filename, "w+") as jsonfile:
        json.dump(json_output, jsonfile, indent=4)
    logger.debug(f"Wrote token data to cache file: {cache_filename}")


def read_collection_data_from_file(expected_supply: int, slug: str) -> list[Token]:
    cache_filename = get_cache_filename(slug)
    tokens = []
    try:
        with open(cache_filename) as jsonfile:
            tokens_data = json.load(jsonfile)
            if len(tokens_data) != expected_supply:
                logger.warning(
                    "Warning: Data cache file for %s collection has data for %s tokens "
                    "but total supply fetched from opensea is %s",
                    slug,
                    len(tokens_data),
                    expected_supply,
                )
            non_null_tokens = 0
            if len(tokens_data) > 0:
                for token_data in tokens_data:
                    if token_data["metadata_dict"]:
                        non_null_tokens += 1
                        tokens.append(Token.from_dict(token_data))
            null_tokens = len(tokens_data) - non_null_tokens
            if null_tokens:
                msg = (
                    f"Warning: Data cache file had empty metadata for {null_tokens} "
                    "tokens. This is expected if those tokens are burned or "
                    "unrevealed. However, they are not taken into account into "
                    "rarity. Please check the cache file for errors."
                )
                logger.warning(msg)
                print(msg)
        logger.debug(f"Read {len(tokens)} tokens from cache file: {cache_filename}")
    except FileNotFoundError:
        logger.warning(f"No opensea cache file found for {slug}: {cache_filename}")
    except Exception:
        logger.exception(
            "Failed to parse valid cache data for %s from %s",
            slug,
            cache_filename,
            exc_info=True,
        )
        return []

    return tokens


# NFT metadata standard type definitions described here:
# https://docs.opensea.io/docs/metadata-standards
def is_string_trait(trait: dict) -> bool:
    """Helper method to verify string trait"""
    return trait[OS_METADATA_TRAIT_TYPE] is None


def is_numeric_trait(trait: dict) -> bool:
    """Helper method to verify numeric trait"""
    return trait[OS_METADATA_TRAIT_TYPE] in [
        "number",
        "boost_percentage",
        "boost_number",
    ]


def is_date_trait(trait: dict) -> bool:
    """Helper method to verify date trait"""
    return trait[OS_METADATA_TRAIT_TYPE] in ["date"]



from dataclasses import dataclass
from enum import Enum

from open_rarity.models.token import Token


class RankProvider(Enum):
    # external ranking providers
    TRAITS_SNIPER = "traits_sniper"
    RARITY_SNIFFER = "rarity_sniffer"
    RARITY_SNIPER = "rarity_sniper"

    # open rarity scoring
    OR_ARITHMETIC = "or_arithmetic"
    OR_GEOMETRIC = "or_geometric"
    OR_HARMONIC = "or_harmonic"
    OR_SUM = "or_sum"
    OR_INFORMATION_CONTENT = "or_information_content"


EXTERNAL_RANK_PROVIDERS = [
    RankProvider.TRAITS_SNIPER,
    RankProvider.RARITY_SNIFFER,
    RankProvider.RARITY_SNIPER,
]

Rank = int
Score = float


@dataclass
class RarityData:
    provider: RankProvider
    rank: Rank
    score: Score | None = None


@dataclass
class TokenWithRarityData:
    token: Token
    rarities: list[RarityData]



import math

from open_rarity.models.collection import Collection
from open_rarity.models.token_rarity import TokenRarity
from open_rarity.scoring.scorer import Scorer
from open_rarity.scoring.token_feature_extractor import TokenFeatureExtractor


class RarityRanker:
    """This class is used to rank a set of tokens given their rarity scores."""

    default_scorer = Scorer()

    @staticmethod
    def rank_collection(
        collection: Collection, scorer: Scorer = default_scorer
    ) -> list[TokenRarity]:
        """
        Ranks tokens in the collection with the default scorer implementation.
        Scores that are higher indicate a higher rarity, and thus a lower rank.

        Tokens with the same score will be assigned the same rank, e.g. we use RANK
        (vs. DENSE_RANK).
        Example: 1, 2, 2, 2, 5.
        Scores are considered the same rank if they are within about 9 decimal digits
        of each other.


        Parameters
        ----------
        collection : Collection
            Collection object with populated tokens
        scorer: Scorer
            Scorer instance

        Returns
        -------
        list[TokenRarity]
            list of TokenRarity objects with score, rank and token information
            sorted by rank
        """

        if (
            collection is None
            or collection.tokens is None
            or len(collection.tokens) == 0
        ):
            return []

        tokens = collection.tokens
        scores: list[float] = scorer.score_tokens(collection, tokens=tokens)

        # fail ranking if dimension of scores doesn't match dimension of tokens
        assert len(tokens) == len(scores)

        token_rarities: list[TokenRarity] = []

        # augment collection tokens with score information
        for idx, token in enumerate(tokens):
            # extract features from the token
            token_features = TokenFeatureExtractor.extract_unique_attribute_count(
                token=token, collection=collection
            )

            token_rarities.append(
                TokenRarity(
                    token=token,
                    score=scores[idx],
                    token_features=token_features,
                )
            )

        return RarityRanker.set_rarity_ranks(token_rarities)

    @staticmethod
    def set_rarity_ranks(
        token_rarities: list[TokenRarity],
    ) -> list[TokenRarity]:
        """Ranks a set of tokens according to OpenRarity algorithm.
        To account for additional factors like unique items in a collection,
        OpenRarity implements multi-factor sort. Current sort algorithm uses two
        factors: unique attributes count and Information Content score, in order.
        Tokens with the same score will be assigned the same rank, e.g. we use RANK
        (vs. DENSE_RANK).
        Example: 1, 2, 2, 2, 5.
        Scores are considered the same rank if they are within about 9 decimal digits
        of each other.

        Parameters
        ----------
        token_rarities : list[TokenRarity]
            unordered list of tokens with rarity score
            information that should have the ranks set on

        Returns
        -------
        list[TokenRarity]
            modified input token_rarities with ranking data set,
            ordered by rank ascending and score descending

        """
        sorted_token_rarities: list[TokenRarity] = sorted(
            token_rarities,
            key=lambda k: (
                k.token_features.unique_attribute_count,
                k.score,
            ),
            reverse=True,
        )

        # Perform ranking of each token in collection
        for i, token_rarity in enumerate(sorted_token_rarities):
            rank = i + 1
            if i > 0:
                prev_token_rarity = sorted_token_rarities[i - 1]
                scores_equal = math.isclose(token_rarity.score, prev_token_rarity.score)
                if scores_equal:
                    assert prev_token_rarity.rank is not None
                    rank = prev_token_rarity.rank

            token_rarity.rank = rank

        return sorted_token_rarities



from dataclasses import dataclass

from open_rarity.models.collection import Collection


@dataclass
class CollectionWithMetadata:
    # This class is just used for resolving different rarity ranks from
    # various providers and allowing us to compare across rarity providers.
    collection: Collection
    contract_addresses: list[str]
    token_total_supply: int
    opensea_slug: str



import logging

import requests

from .rank_resolver import RankResolver

logger = logging.getLogger("open_rarity_logger")
RARITY_SNIFFER_API_URL = "https://raritysniffer.com/api/index.php"
USER_AGENT = {
    "User-Agent": (
        "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 "
        "(KHTML, like Gecko) Chrome/96.0.4664.45 Safari/537.36"  # noqa: E501
    )
}


class RaritySnifferResolver(RankResolver):
    @staticmethod
    def get_all_ranks(contract_address: str) -> dict[str, int]:
        """Fetches all available tokens and ranks
        for a given collection from rarity sniffer.
        Only usable for EVM tokens and collections for a single
        contract address.

        Parameters
        ----------
        contract_address : The contract address of the collection

        Returns
        -------
        dict[int, int]: Dictionary of token ID # to the rank. Empty if no data.

        Raises
        ------
        Exception
            If call to the rarity sniffer failed the method throws exception
        """
        querystring = {
            "query": "fetch",
            "collection": contract_address,
            "taskId": "any",
            "norm": "true",
            "partial": "false",
            "traitCount": "true",
        }

        response = requests.request(
            "GET",
            RARITY_SNIFFER_API_URL,
            params=querystring,
            headers=USER_AGENT,
        )

        if response.status_code != 200:
            logger.debug(
                "[RaritySniffer] Failed to resolve Rarity Sniffer ranks for "
                f"{contract_address}. Received: {response.status_code}: "
                f"{response.reason} {response.json()}"
            )
            response.raise_for_status()

        data = response.json().get("data", None)
        if "Not found" in response.json().get("error", "") or not data:
            logger.exception(
                f"[RaritySniffer] No data for {contract_address}. "
                f"Json response: {response.json()}",
                exc_info=True,
            )
            return {}
        try:
            token_ids_to_ranks = {
                str(nft["id"]): int(nft["positionId"]) for nft in data
            }
        except Exception:
            logger.exception(
                f"[RaritySniffer] Data could not be parsed for {contract_address}. "
                f"Json response: {response.json()}",
                exc_info=True,
            )
            return {}

        return token_ids_to_ranks



from .collection import Collection
from .token import Token
from .token_identifier import EVMContractTokenIdentifier
from .token_metadata import StringAttribute, TokenMetadata
from .token_rarity import TokenRarity
from .token_standard import TokenStandard



from open_rarity.models.collection import Collection
from open_rarity.models.token import Token
from open_rarity.models.token_standard import TokenStandard
from open_rarity.scoring.handlers.information_content_scoring_handler import (
    InformationContentScoringHandler,
)
from open_rarity.scoring.scoring_handler import ScoringHandler


class Scorer:
    """Scorer is the main class to score rarity scores for a given
    collection and token(s) based on the default OpenRarity scoring
    algorithm.
    """

    handler: ScoringHandler

    def __init__(self) -> None:
        # OpenRarity uses InformationContent as the scoring algorithm of choice.
        self.handler = InformationContentScoringHandler()

    def validate_collection(self, collection: Collection) -> None:
        """Validate collection eligibility for OpenRarity scoring

        Parameters
        ----------
        collection: Collection)
            The collection to validate
        """
        if collection.has_numeric_attribute:
            raise ValueError(
                "OpenRarity currently does not support collections with "
                "numeric or date traits"
            )

        allowed_standards = {
            TokenStandard.ERC721,
            TokenStandard.METAPLEX_NON_FUNGIBLE,
        }

        if not set(collection.token_standards).issubset(allowed_standards):
            raise ValueError(
                "OpenRarity currently only supports ERC721/Non-fungible standards"
            )

    def score_token(self, collection: Collection, token: Token) -> float:
        """Scores an individual token based on the traits distribution across
        the whole collection.

        Parameters
        ----------
        collection : Collection
            The collection to score from
        token : Token
            a single Token to score

        Returns
        -------
        float
            The score of the token
        """
        self.validate_collection(collection=collection)
        return self.handler.score_token(collection=collection, token=token)

    def score_tokens(self, collection: Collection, tokens: list[Token]) -> list[float]:
        """Used if you only want to score a batch of tokens that belong to collection.
        This will typically be more efficient than calling score_token for each
        token in `tokens`.

        Parameters
        ----------
        collection : Collection
            The collection to score from
        tokens : list[Token]
            a batch of tokens belonging to collection to be scored

        Returns
        -------
        list[float]
            list of scores in order of `tokens`
        """
        self.validate_collection(collection=collection)
        return self.handler.score_tokens(collection=collection, tokens=tokens)

    def score_collection(self, collection: Collection) -> list[float]:
        """Scores all tokens on collection.tokens

        Parameters
        ----------
        collection : Collection
            The collection to score all tokens from

        Returns
        -------
        list[float]
            list of scores in order of `collection.tokens`
        """
        self.validate_collection(collection=collection)
        return self.handler.score_tokens(
            collection=collection,
            tokens=collection.tokens,
        )

    def score_collections(self, collections: list[Collection]) -> list[list[float]]:
        """Scores all tokens in every collection provided.

        Parameters
        ----------
        collections: list[Collection])
            The collections to score

        Returns
        -------
        list[list[float]]
            A list of scores for all tokens in each given Collection,
            ordered by the collection's `tokens` field.
        """
        for collection in collections:
            self.validate_collection(collection=collection)
        return [
            self.handler.score_tokens(collection=c, tokens=c.tokens)
            for c in collections
        ]



from enum import Enum


class TokenStandard(Enum):
    """Enum class representing the interface or standard that
    a token is respecting. Each chain may have their own token standards.
    """

    # -- Ethereum/EVM standards
    # https://eips.ethereum.org/EIPS/eip-721
    ERC721 = "erc721"
    # https://eips.ethereum.org/EIPS/eip-1155
    ERC1155 = "erc1155"

    # -- Solana token standards
    # https://docs.metaplex.com/programs/token-metadata/token-standard
    METAPLEX_NON_FUNGIBLE = "metaplex_non_fungible"



from .models import (
    Collection,
    EVMContractTokenIdentifier,
    StringAttribute,
    Token,
    TokenMetadata,
    TokenRarity,
    TokenStandard,
)
from .rarity_ranker import RarityRanker
from .scoring import Scorer as OpenRarityScorer



from dataclasses import dataclass

from open_rarity.models.token import Token
from open_rarity.models.token_ranking_features import TokenRankingFeatures


@dataclass
class TokenRarity:
    """The class holds rarity and optional rank information along with the token

    Attributes
    ----------
    score : float
        OpenRarity score for the token within collection
    token : Token
        token class
    token_features : TokenFeatures
        various token features
    rank: int | None
        rank of the token within the collection
    """

    score: float
    token_features: TokenRankingFeatures
    token: Token
    rank: int | None = None



import logging
import os
import time

import requests

from .rank_resolver import RankResolver

logger = logging.getLogger("open_rarity_logger")
# For fetching rank fetches for an entire contract
TRAIT_SNIPER_RANKS_URL = (
    "https://api.traitsniper.com/v1/collections/{contract_address}/ranks"
)
# For single rank fetches
TRAIT_SNIPER_NFTS_URL = "https://api.traitsniper.com/api/projects/{slug}/nfts"

USER_AGENT = {
    "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/96.0.4664.45 Safari/537.36"  # noqa: E501
}
API_KEY = os.environ.get("TRAIT_SNIPER_API_KEY") or ""


class TraitSniperResolver(RankResolver):
    @staticmethod
    def get_all_ranks(contract_address: str) -> dict[str, int]:
        """Get all ranks for a contract address

        Returns
        -------
        dict[str, int]
            A dictionary of token_id to ranks
        """
        rank_data_page = TraitSniperResolver.get_ranks(contract_address, page=1)
        all_rank_data = rank_data_page
        page = 2

        while rank_data_page:
            rank_data_page = TraitSniperResolver.get_ranks(contract_address, page=page)
            all_rank_data.extend(rank_data_page)
            page += 1
            # Due to rate limits we need to slow things down a bit...
            # Free tier is 5 request per second
            time.sleep(12)

        return {
            str(rank_data["token_id"]): int(rank_data["rarity_rank"])
            for rank_data in all_rank_data
            if rank_data["rarity_rank"]
        }

    @staticmethod
    def get_ranks(contract_address: str, page: int, limit: int = 200) -> list[dict]:
        """
        Parameters
        ----------
        contract_address: str
            The contract address of collection you're fetching ranks for
        limit: int
            The number of ranks to fetch. Defaults to 200, and maxes at 200
            due to API limitations.

        Returns
        -------
            List of rarity rank data from trait sniper API with the following
            data structure for each item in the list:
            {
                "rarity_rank": int,
                "rarity_score": float
                "token_id": str
            }

        Requires
        --------
            API KEY

        Raises
        ------
            ValueError if contract address is None
        """
        if not contract_address:
            msg = f"Failed to fetch traitsniper. {contract_address=} is invalid."
            logger.exception(msg)
            raise ValueError(msg)

        url = TRAIT_SNIPER_RANKS_URL.format(contract_address=contract_address)
        headers = {
            **USER_AGENT,
            **{"X-TS-API-KEY": API_KEY},
        }
        query_params = {
            "limit": max(limit, 200),
            "page": page,
        }
        response = requests.request("GET", url, headers=headers, params=query_params)
        if response.status_code == 200:
            return response.json()["ranks"]
        else:
            if (
                "Collection could not be found on TraitSniper"
                in response.json()["message"]
            ):
                logger.warning(
                    f"[TraitSniper] Collection not found: {contract_address}"
                )
            else:
                logger.debug(
                    "[TraitSniper] Failed to resolve TraitSniper rank for "
                    f"{contract_address}. Received {response.status_code} "
                    f"for {url}: {response.reason}. {response.json()}"
                )
            return []

    @staticmethod
    def get_rank(collection_slug: str, token_id: int) -> int | None:
        """Sends a GET request to Trait Sniper API to fetch ranking
        data for a given EVM token. Trait Sniper uses opensea slug as a param.

        Parameters
        ----------
        collection_slug : str
            collection slug of collection you're attempting to fetch. This must be
            the slug on trait sniper's slug system.
        token_id : int
            the token number.

        Returns
        -------
        int | None
            Rarity rank for given token ID if request was successful, otherwise None.

        Raises
        ------
        ValueError
            If slug is invalid.
        """
        # TODO [vicky]: In future, we can add retry mechanisms if needed

        querystring: dict[str, str | int] = {
            "trait_norm": "true",
            "trait_count": "true",
            "token_id": token_id,
        }

        if not collection_slug:
            msg = "Cannot fetch traitsniper rank as slug is None."
            logger.exception(msg)
            raise ValueError(msg)

        url = TRAIT_SNIPER_NFTS_URL.format(slug=collection_slug)
        response = requests.request("GET", url, params=querystring, headers=USER_AGENT)
        if response.status_code == 200:
            return int(response.json()["nfts"][0]["rarity_rank"])
        else:
            logger.debug(
                "[TraitSniper] Failed to resolve TraitSniper rank for "
                f"{collection_slug} {token_id}. Received {response.status_code} "
                f"for {url}: {response.reason}. {response.json()}"
            )
            return None



import warnings
from collections import defaultdict
from dataclasses import dataclass
from functools import cached_property

from open_rarity.models.token import Token
from open_rarity.models.token_metadata import (
    AttributeName,
    AttributeValue,
    StringAttribute,
)
from open_rarity.models.token_standard import TokenStandard
from open_rarity.models.utils.attribute_utils import normalize_attribute_string

TRAIT_COUNT_ATTRIBUTE_NAME = "meta_trait:trait_count"


@dataclass
class CollectionAttribute:
    """Class represents an attribute that at least one token in a Collection has.
    E.g. "hat" = "cap" would be one attribute, and "hat" = "beanie" would be another
    unique attribute, even though they may belong to the same attribute type (id=name).

    Attributes
    ----------
    attribute : StringAttribute
        the unique attribute pair
    total_tokens : int
        total number of tokens in the collection that have this attribute
    """

    attribute: StringAttribute
    total_tokens: int


@dataclass
class Collection:
    """Class represents collection of tokens used to determine token rarity score.
    A token's rarity is influenced by the attribute frequency of all the tokens
    in a collection.

    Attributes
    ----------
    tokens : list[Token]
        list of all Tokens that belong to the collection
    attributes_frequency_counts: dict[AttributeName, dict[AttributeValue, int]]
        dictionary of attributes to the number of tokens in this collection that has
        a specific value for every possible value for the given attribute.

        If not provided, the attributes distribution will be derived from the
        attributes on the tokens provided.

        Example:
            {"hair": {"brown": 500, "blonde": 100}
            which means 500 tokens has hair=brown, 100 token has hair=blonde

        Note: All trait names and string values should be lowercased and stripped
        of leading and trialing whitespace.
        Note 2: We currently only support string attributes.

    name: A reference string only used for debugger log lines

    We do not recommend resetting @tokens attribute after Collection initialization
    as that will mess up cached property values:
        has_numeric_attributes
        get_token_standards
    """

    attributes_frequency_counts: dict[AttributeName, dict[AttributeValue, int]]
    name: str

    def __init__(
        self,
        tokens: list[Token],
        # Deprecated - Kept to not break interface, but is not used.
        # We always coimpute the attributes_frequency_counts from the tokens to avoid
        # divergence.
        # TODO [10/16/22]: To remove in 1.0 release
        attributes_frequency_counts: dict[AttributeName, dict[AttributeValue, int]]
        | None = None,
        name: str | None = "",
    ):
        if attributes_frequency_counts is not None:
            warnings.warn(
                "`attribute_frequency_counts` is deprecated and will be removed. "
                "Counts will be derived from the token data.",
                DeprecationWarning,
                stacklevel=2,
            )
        self._trait_countify(tokens)
        self._tokens = tokens
        self.name = name or ""
        self.attributes_frequency_counts = (
            self._derive_normalized_attributes_frequency_counts()
        )

    @property
    def tokens(self) -> list[Token]:
        return self._tokens

    @property
    def token_total_supply(self) -> int:
        return len(self._tokens)

    @cached_property
    def has_numeric_attribute(self) -> bool:
        return (
            next(
                filter(
                    lambda t: len(t.metadata.numeric_attributes)
                    or len(t.metadata.date_attributes),
                    self._tokens,
                ),
                None,
            )
            is not None
        )

    @cached_property
    def token_standards(self) -> list[TokenStandard]:
        """Returns token standards for this collection.

        Returns
        -------
        list[TokenStandard]
            the set of unique token standards that any token in this collection
            interfaces or uses.
        """
        token_standards = set()
        for token in self._tokens:
            token_standards.add(token.token_standard)
        return list(token_standards)

    def total_tokens_with_attribute(self, attribute: StringAttribute) -> int:
        """Returns the numbers of tokens in this collection with the attribute
        based on the attributes frequency counts.

        Returns
        -------
        int
            The number of tokens with attribute (attribute_name, attribute_value)
        """
        return self.attributes_frequency_counts.get(attribute.name, {}).get(
            attribute.value, 0
        )

    def total_attribute_values(self, attribute_name: str) -> int:
        return len(self.attributes_frequency_counts.get(attribute_name, {}))

    def extract_null_attributes(
        self,
    ) -> dict[AttributeName, CollectionAttribute]:
        """Compute probabilities of Null attributes.

        Returns
        -------
        dict[AttributeName(str), CollectionAttribute(str)]
            dict of attribute name to the number of assets without the attribute
            (e.g. # of assets where AttributeName=NULL)
        """
        result = {}
        for (
            trait_name,
            trait_values,
        ) in self.attributes_frequency_counts.items():
            # To obtain probabilities for missing attributes
            # e.g. value of trait not set for the asset
            #
            # We sum all values counts for particular
            # attributes and subtract it from total supply.
            # This number divided by total supply is a
            # probability of Null attribute
            total_trait_count = sum(trait_values.values())

            # compute null trait probability
            # only if there is a positive number of assets without
            # this trait
            assets_without_trait = self.token_total_supply - total_trait_count
            if assets_without_trait > 0:
                result[trait_name] = CollectionAttribute(
                    attribute=StringAttribute(trait_name, "Null"),
                    total_tokens=assets_without_trait,
                )

        return result

    def extract_collection_attributes(
        self,
    ) -> dict[AttributeName, list[CollectionAttribute]]:
        """Extracts the map of collection traits with it's respective counts

        Returns
        -------
        dict[str, CollectionAttribute]
            dict of attribute name to count of assets missing the attribute
        """

        collection_traits: dict[str, list[CollectionAttribute]] = defaultdict(list)

        for (
            trait_name,
            trait_value_dict,
        ) in self.attributes_frequency_counts.items():
            for trait_value, trait_count in trait_value_dict.items():
                collection_traits[trait_name].append(
                    CollectionAttribute(
                        attribute=StringAttribute(trait_name, str(trait_value)),
                        total_tokens=trait_count,
                    )
                )

        return collection_traits

    def _trait_countify(self, tokens: list[Token]) -> None:
        """Updates tokens to have meta attribute "meta trait: trait_count" if it doesn't
        already exist.

        Parameters
        ----------
        tokens : list[Token]
            List of tokens to add trait count attribute to. Modifies in place.

        """
        for token in tokens:
            trait_count = token.trait_count()
            if token.has_attribute(TRAIT_COUNT_ATTRIBUTE_NAME):
                trait_count -= 1
            # NOTE: There is a chance we override an existing attribute here, but it's
            # highly unlikely that a token would have a trait_count attribute to begin
            # with (no known collections have it right now).
            # To decrease the chance of collision, we pre-pend "meta trait: ".
            # If an existing trait count attribute already exists with a different name,
            # we will not remove it. In the future, we can refactor to distinguish
            # between meta and non-meta attributes.
            token.metadata.add_attribute(
                StringAttribute(name=TRAIT_COUNT_ATTRIBUTE_NAME, value=str(trait_count))
            )

    def _normalize_attributes_frequency_counts(
        self,
        attributes_frequency_counts: dict[AttributeName, dict[AttributeValue, int]],
    ) -> dict[AttributeName, dict[AttributeValue, int]]:
        """We normalize all collection attributes to ensure that neither casing nor
        leading/trailing spaces produce different attributes:
            (e.g. 'Hat' == 'hat' == 'hat ')
        If a collection has the following in their attributes frequency counts:
            ('Hat', 'beanie') 5 tokens and
            ('hat', 'beanie') 10 tokens
        this would produce: ('hat', 'beanie') 15 tokens
        """
        normalized: dict[AttributeName, dict[AttributeValue, int]] = {}
        for (
            attr_name,
            attr_value_to_count,
        ) in attributes_frequency_counts.items():
            normalized_name = normalize_attribute_string(attr_name)
            if normalized_name not in normalized:
                normalized[normalized_name] = {}
            for attr_value, attr_count in attr_value_to_count.items():
                normalized_value = (
                    normalize_attribute_string(attr_value)
                    if isinstance(attr_value, str)
                    else attr_value
                )
                if normalized_value not in normalized[normalized_name]:
                    normalized[normalized_name][normalized_value] = attr_count
                else:
                    normalized[normalized_name][normalized_value] += attr_count

        return normalized

    def _derive_normalized_attributes_frequency_counts(
        self,
    ) -> dict[AttributeName, dict[AttributeValue, int]]:
        """Derives and constructs attributes_frequency_counts based on
        string attributes on tokens. Numeric or date attributes currently not
        supported.

        Returns
        -------
        dict[ AttributeName, dict[AttributeValue, int] ]
            dictionary of attributes to the number of tokens in this collection
            that has a specific value for every possible value for the given
            attribute, by default None.
        """
        attrs_freq_counts: dict[AttributeName, dict[AttributeValue, int]] = defaultdict(
            dict
        )

        for token in self._tokens:
            for (
                attr_name,
                str_attr,
            ) in token.metadata.string_attributes.items():
                normalized_name = normalize_attribute_string(attr_name)
                if str_attr.value not in attrs_freq_counts[attr_name]:
                    attrs_freq_counts[normalized_name][str_attr.value] = 1
                else:
                    attrs_freq_counts[normalized_name][str_attr.value] += 1

        return dict(attrs_freq_counts)

    def __str__(self) -> str:
        return f"Collection[{self.name}]"



from dataclasses import dataclass
from typing import Any, Iterable

from open_rarity.models.token_identifier import (
    EVMContractTokenIdentifier,
    SolanaMintAddressTokenIdentifier,
    TokenIdentifier,
    get_identifier_class_from_dict,
)
from open_rarity.models.token_metadata import (
    Attribute,
    AttributeName,
    StringAttribute,
    TokenMetadata,
)
from open_rarity.models.token_standard import TokenStandard
from open_rarity.models.utils.attribute_utils import normalize_attribute_string


@dataclass
class Token:
    """Class represents a token on the blockchain.
    Examples of these are non-fungible tokens, or semi-fungible tokens.

    Attributes
    ----------
    token_identifier : TokenIdentifier
        data representing how the token is identified, which may be based
        on the token_standard or chain it lives on.
    token_standard : TokenStandard
        name of token standard (e.g. EIP-721 or EIP-1155)
    metadata: TokenMetadata
        contains the metadata of this specific token
    """

    token_identifier: TokenIdentifier
    token_standard: TokenStandard
    metadata: TokenMetadata

    @classmethod
    def from_erc721(
        cls,
        contract_address: str,
        token_id: int,
        metadata_dict: dict[AttributeName, Any],
    ):
        """Creates a Token class representing an ERC721 evm token given the following
        parameters.

        Parameters
        ----------
        contract_address : str
            Contract address of the token
        token_id : int
            Token ID number of the token
        metadata_dict : dict
            Dictionary of attribute name to attribute value for the given token.
            The type of the value determines whether the attribute is a string,
            numeric or date attribute.

            class           attribute type
            ------------    -------------
            string          string attribute
            int | float     numeric_attribute
            datetime        date_attribute (stored as timestamp, seconds from epoch)

        Returns
        -------
        Token
            A Token instance with EVMContractTokenIdentifier and ERC721 standard set.
        """
        return cls(
            token_identifier=EVMContractTokenIdentifier(
                contract_address=contract_address, token_id=token_id
            ),
            token_standard=TokenStandard.ERC721,
            metadata=TokenMetadata.from_attributes(metadata_dict),
        )

    @classmethod
    def from_metaplex_non_fungible(
        cls, mint_address: str, attributes: dict[AttributeName, Any]
    ):
        """Creates a Token class representing a Metaplex non-fungible token
        given the following parameters.

        Parameters
        ----------
        mint_address: str
            The mint address of the token.
        attributes : dict
            Dictionary of attribute name to attribute value for the given token.
            Same as the attributes in from_erc721.

        Returns
        -------
        Token
            A Token instance with SolanaMintAddressTokenIdentifier and
            METAPLEX_NON_FUNGIBLE standard set.
        """
        return cls(
            token_identifier=SolanaMintAddressTokenIdentifier(
                mint_address=mint_address,
            ),
            token_standard=TokenStandard.METAPLEX_NON_FUNGIBLE,
            metadata=TokenMetadata.from_attributes(attributes),
        )

    @classmethod
    def from_dict(cls, data_dict: dict):
        identifier_class = get_identifier_class_from_dict(data_dict["token_identifier"])

        return cls(
            token_identifier=identifier_class.from_dict(data_dict["token_identifier"]),
            token_standard=TokenStandard[data_dict["token_standard"]],
            metadata=TokenMetadata.from_attributes(data_dict["metadata_dict"]),
        )

    def attributes(self) -> dict[AttributeName, Any]:
        return self.metadata.to_attributes()

    def has_attribute(self, attribute_name: str) -> bool:
        return self.metadata.attribute_exists(attribute_name)

    def trait_count(self) -> int:
        """Returns the count of non-null, non-"none" value traits this token has."""

        def get_attributes_count(attributes: Iterable[Attribute]) -> int:
            return sum(
                map(
                    lambda a: (
                        not isinstance(a, StringAttribute)
                        or normalize_attribute_string(a.value) not in ("none", "")
                    ),
                    attributes,
                )
            )

        return (
            get_attributes_count(self.metadata.string_attributes.values())
            + get_attributes_count(self.metadata.numeric_attributes.values())
            + get_attributes_count(self.metadata.date_attributes.values())
        )

    def to_dict(self) -> dict:
        return {
            "token_identifier": self.token_identifier.to_dict(),
            "metadata_dict": self.attributes(),
            "token_standard": self.token_standard.name,
        }

    def __str__(self):
        return f"Token[{self.token_identifier}]"



def normalize_attribute_string(value: str) -> str:
    """Normalizes either attribute names or string attribute values.
    This is a helper function to ensure we are consistently normalizing
    by always lower casing and stripping input string.

    Parameters
    ----------
    value : str
        The string to normalize
        (this should be either attribute name or a string attribute value)

    Returns
    -------
    str
        normalized string
    """
    return value.lower().strip()



from abc import abstractmethod
from typing import Protocol

from open_rarity.models.collection import Collection
from open_rarity.models.token import Token


class ScoringHandler(Protocol):
    """ScoringHandler class is an interface for different scoring algorithms to
    implement. Sub-classes are responsibile to ensure the batch functions are
    efficient for their particular algorithm.
    """

    @abstractmethod
    def score_token(self, collection: Collection, token: Token) -> float:
        """Scores an individual token based on the traits distribution across
        the whole collection.

        Parameters
        ----------
        collection : Collection
            The collection with the attributes frequency counts to base the
            token trait probabilities on to calculate score.
        token : Token
            The token to score

        Returns
        -------
        float
            The token score
        """
        raise NotImplementedError

    @abstractmethod
    def score_tokens(
        self,
        collection: Collection,
        tokens: list[Token],
    ) -> list[float]:
        """Used if you only want to score a batch of tokens that belong to collection.
        This will typically be more efficient than calling score_token for each
        token in `tokens`.

        Parameters
        ----------
        collection : Collection
            The collection to score from
        tokens : list[Token]
            a batch of tokens belonging to collection to be scored

        Returns
        -------
        list[float]
            list of scores in order of `tokens`
        """
        raise NotImplementedError



from open_rarity.models import Token
from open_rarity.models.collection import Collection
from open_rarity.models.token_ranking_features import TokenRankingFeatures


class TokenFeatureExtractor:
    """
    Utility class that extract features from tokens
    """

    @staticmethod
    def extract_unique_attribute_count(
        token: Token, collection: Collection
    ) -> TokenRankingFeatures:
        """This method extracts unique attributes count from the token

        Parameters
        ----------
        token : Token
            The token to extract features from
        collection : Collection
            The collection with the attributes frequency counts to base the
            token trait probabilities on to calculate score.

        Returns
        -------
        TokenFeatures
            Token features wrapper class that contains extracted features

        """

        unique_attributes_count: int = 0

        for string_attribute in token.metadata.string_attributes.values():
            count = collection.total_tokens_with_attribute(string_attribute)

            if count == 1:
                unique_attributes_count += 1

        return TokenRankingFeatures(unique_attribute_count=unique_attributes_count)



import datetime
from dataclasses import dataclass, field
from typing import Any

from open_rarity.models.utils.attribute_utils import normalize_attribute_string

AttributeName = str
AttributeValue = str


@dataclass
class StringAttribute:
    """Class represents string token attribute name and value

    Attributes
    ----------
    name : AttributeName
        name of an attribute
    value : str
        value of a string attribute
    """

    name: AttributeName  # duplicate name here for ease of reduce
    value: AttributeValue

    def __init__(self, name: AttributeName, value: AttributeValue):
        # We treat string attributes name and value the same regardless of
        # casing or leading/trailing whitespaces.
        self.name = normalize_attribute_string(name)
        self.value = normalize_attribute_string(value)


@dataclass
class NumericAttribute:
    """Class represents numeric token attribute name and value

    Attributes
    ----------
    name : AttributeName
        name of an attribute
    value : float | int
        value of a numeric attribute
    """

    name: AttributeName
    value: float | int

    def __init__(self, name: AttributeName, value: float | int):
        # We treat attributes names the same regardless of
        # casing or leading/trailing whitespaces.
        self.name = normalize_attribute_string(name)
        self.value = value


@dataclass
class DateAttribute:
    """Class represents date token attribute name and value

    Attributes
    ----------
    name : AttributeName
        name of an attribute
    value : int
        value of a date attribute in UNIX timestamp format
    """

    name: AttributeName
    value: int

    def __init__(self, name: AttributeName, value: int):
        # We treat attributes names the same regardless of
        # casing or leading/trailing whitespaces.
        self.name = normalize_attribute_string(name)
        self.value = value


Attribute = StringAttribute | NumericAttribute | DateAttribute


@dataclass
class TokenMetadata:
    """Class represents EIP-721 or EIP-1115 compatible metadata structure

    Attributes
    ----------
    string_attributes : dict
        mapping of atrribute name to list of string attribute values
    numeric_attributes : dict
        mapping of atrribute name to list of numeric attribute values
    date_attributes : dict
        mapping of attribute name to list of date attribute values


    All attributes names are normalized and all string attribute values are
    normalized in the same way - lowercased and leading/trailing whitespace stripped.
    """

    string_attributes: dict[AttributeName, StringAttribute] = field(
        default_factory=dict
    )
    numeric_attributes: dict[AttributeName, NumericAttribute] = field(
        default_factory=dict
    )
    date_attributes: dict[AttributeName, DateAttribute] = field(default_factory=dict)

    def __post_init__(self):
        self.string_attributes = self._normalize_attributes_dict(self.string_attributes)
        self.numeric_attributes = self._normalize_attributes_dict(
            self.numeric_attributes
        )
        self.date_attributes = self._normalize_attributes_dict(self.date_attributes)

    @classmethod
    def from_attributes(cls, attributes: dict[AttributeName, Any]):
        """Constructs TokenMetadata class based on an attributes dictionary

        Parameters
        ----------
        attributes : dict[AttributeName, Any]
            Dictionary of attribute name to attribute value for the given token.
            The type of the value determines whether the attribute is a string,
            numeric or date attribute.

            class           attribute type
            ------------    -------------
            string          string attribute
            int | float     numeric_attribute
            datetime        date_attribute (stored as timestamp, seconds from epoch)

        Returns
        -------
        TokenMetadata
            token metadata from input
        """
        string_attributes = {}
        numeric_attributes = {}
        date_attributes = {}
        for attr_name, attr_value in attributes.items():
            if isinstance(attr_value, str):
                string_attributes[attr_name] = StringAttribute(
                    name=attr_name, value=attr_value
                )
            elif isinstance(attr_value, (float, int)):
                numeric_attributes[attr_name] = NumericAttribute(
                    name=attr_name, value=attr_value
                )
            elif isinstance(attr_value, datetime.datetime):
                date_attributes[attr_name] = DateAttribute(
                    name=attr_name,
                    value=int(attr_value.timestamp()),
                )
            else:
                raise TypeError(
                    f"Provided attribute value has invalid type: {type(attr_value)}. "
                    "Must be either str, float, int or datetime."
                )

        return cls(
            string_attributes=string_attributes,
            numeric_attributes=numeric_attributes,
            date_attributes=date_attributes,
        )

    def to_attributes(self) -> dict[AttributeName, Any]:
        """Returns a dictionary of all attributes in this metadata object."""
        attributes: dict[AttributeName, Any] = {}
        for str_attr in self.string_attributes.values():
            attributes[str_attr.name] = str_attr.value
        for num_attr in self.numeric_attributes.values():
            attributes[num_attr.name] = num_attr.value
        for date_attr in self.date_attributes.values():
            attributes[date_attr.name] = datetime.datetime.fromtimestamp(
                date_attr.value
            )
        return attributes

    def add_attribute(self, attribute: Attribute):
        """Adds an attribute to this metadata object, overriding existing
        attribute if the normalized attribute name already exists."""
        if isinstance(attribute, StringAttribute):
            self.string_attributes[attribute.name] = attribute
        elif isinstance(attribute, NumericAttribute):
            self.numeric_attributes[attribute.name] = attribute
        elif isinstance(attribute, DateAttribute):
            self.date_attributes[attribute.name] = attribute
        else:
            raise TypeError(
                f"Provided attribute has invalid type: {type(attribute)}. "
                "Must be either StringAttribute, NumericAttribute or DateAttribute."
            )

    def attribute_exists(self, attribute_name: str) -> bool:
        """Returns True if this metadata object has an attribute with the given name."""
        attr_name = normalize_attribute_string(attribute_name)
        return (
            attr_name in self.string_attributes
            or attr_name in self.numeric_attributes
            or attr_name in self.date_attributes
        )

    def _normalize_attributes_dict(self, attributes_dict: dict) -> dict:
        """Helper function that takes in an attributes dictionary
        and normalizes attribute name in the dictionary to ensure all
        letters are lower cases and whitespace is stripped.
        """
        normalized_attributes_dict = {}
        for attribute_name, attr in attributes_dict.items():
            normalized_attr_name = normalize_attribute_string(attribute_name)
            normalized_attributes_dict[normalized_attr_name] = attr
            if normalized_attr_name != attr.name:
                attr.name = normalized_attr_name
        return normalized_attributes_dict



from abc import abstractmethod
from typing import Protocol, runtime_checkable


@runtime_checkable
class RankResolver(Protocol):
    @staticmethod
    @abstractmethod
    def get_all_ranks(contract_address_or_slug: str) -> dict[str, int]:
        raise NotImplementedError



import logging

import numpy as np

from open_rarity.models.collection import Collection, CollectionAttribute
from open_rarity.models.token import Token
from open_rarity.models.token_metadata import AttributeName
from open_rarity.scoring.utils import get_token_attributes_scores_and_weights

logger = logging.getLogger("open_rarity_logger")


class InformationContentScoringHandler:
    """Rarity describes the information-theoretic "rarity" of a Collection.
    The concept of "rarity" can be considered as a measure of "surprise" at the
    occurrence of a particular token's properties, within the context of the
    Collection from which it is derived. Self-information is a measure of such
    surprise, and information entropy a measure of the expected value of
    self-information across a distribution (i.e. across a Collection).

    It is trivial to "stuff" a Collection with extra information by merely adding
    additional properties to all tokens. This is reflected in the Entropy field,
    measured in bits—all else held equal, a Collection with more token properties
    will have higher Entropy. However, this information bloat is carried by the
    tokens themselves, so their individual information-content grows in line with
    Collection-wide Entropy. The Scores are therefore scaled down by the Entropy
    to provide unitless "relative surprise", which can be safely compared between
    Collections.

    This class computes rarity of each token in the Collection based on information
    entropy. Every TraitType is considered as a categorical probability
    distribution with each TraitValue having an associated probability and hence
    information content. The rarity of a particular token is the sum of
    information content carried by each of its Attributes, divided by the entropy
    of the Collection as a whole (see the Rarity struct for rationale).

    Notably, the lack of a TraitType is considered as a null-Value Attribute as
    the absence across the majority of a Collection implies rarity in those
    tokens that do carry the TraitType.
    """

    # TODO [@danmeshkov]: To support numeric types in a follow-up version.

    def score_token(self, collection: Collection, token: Token) -> float:
        """See ScoringHandler interface.

        Limitations
        -----------
            Does not take into account non-String attributes during scoring.

        """
        return self._score_token(collection, token)

    def score_tokens(
        self,
        collection: Collection,
        tokens: list[Token],
    ) -> list[float]:
        """See ScoringHandler interface.

        Limitations
        -----------
            Does not take into account non-String attributes during scoring.

        """
        # Precompute for performance
        collection_null_attributes = collection.extract_null_attributes()
        collection_attributes = collection.extract_collection_attributes()
        collection_entropy = self._get_collection_entropy(
            collection=collection,
            collection_attributes=collection_attributes,
            collection_null_attributes=collection_null_attributes,
        )
        return [
            self._score_token(
                collection=collection,
                token=t,
                collection_null_attributes=collection_null_attributes,
                # covering the corner case when collection has one item.
                collection_entropy_normalization=collection_entropy
                if collection_entropy
                else 1,
            )
            for t in tokens
        ]

    # Private methods
    def _score_token(
        self,
        collection: Collection,
        token: Token,
        collection_null_attributes: dict[AttributeName, CollectionAttribute] = None,
        collection_entropy_normalization: float = None,
    ) -> float:
        """Calculates the score of the token using information entropy with a
        collection entropy normalization factor.

        Parameters
        ----------
        collection : Collection
            The collection with the attributes frequency counts to base the
            token trait probabilities on to calculate score.
        token : Token
            The token to score
        collection_null_attributes : dict[AttributeName, CollectionAttribute], optional
            Optional memoization of collection.extract_null_attributes(),
            by default None.
        collection_entropy_normalization : float, optional
            Optional memoization of the collection entropy normalization factor,
            by default None.

        Returns
        -------
        float
            The token score
        """
        logger.debug("Computing score for token %s", token)

        ic_token_score = self._get_ic_score(
            collection, token, collection_null_attributes=collection_null_attributes
        )
        logger.debug("IC token score %s", ic_token_score)

        # Now, calculate the collection entropy to use as a normalization for
        # the token score if its not provided
        if collection_entropy_normalization is None:
            collection_entropy = self._get_collection_entropy(
                collection=collection,
                collection_attributes=collection.extract_collection_attributes(),
                collection_null_attributes=collection_null_attributes,
            )
        else:
            collection_entropy = collection_entropy_normalization
        normalized_token_score = ic_token_score / collection_entropy
        logger.debug(
            "Finished scoring %s %s: collection entropy: %s token scores: %s",
            collection,
            token,
            collection_entropy,
            normalized_token_score,
        )

        return normalized_token_score

    def _get_ic_score(
        self,
        collection: Collection,
        token: Token,
        collection_null_attributes: dict[AttributeName, CollectionAttribute] = None,
    ) -> float:
        # First calculate the individual attribute scores for all attributes
        # of the provided token. Scores are the inverted probabilities of the
        # attribute in the collection.
        attr_scores, _ = get_token_attributes_scores_and_weights(
            collection=collection,
            token=token,
            normalized=False,
            collection_null_attributes=collection_null_attributes,
        )

        # Get a single score (via information content) for the token by taking
        # the sum of the logarithms of the attributes' scores.
        return -np.sum(np.log2(np.reciprocal(attr_scores)))

    def _get_collection_entropy(
        self,
        collection: Collection,
        collection_attributes: dict[AttributeName, list[CollectionAttribute]] = None,
        collection_null_attributes: dict[AttributeName, CollectionAttribute] = None,
    ) -> float:
        """Calculates the entropy of the collection, defined to be the
        sum of the probability of every possible attribute name/value pair that
        occurs in the collection times that square root of such probability.

        Parameters
        ----------
        collection : Collection
            The collection to calculate probability on
        collection_attributes : dict[AttributeName, list[CollectionAttribute]], optional
            Optional memoization of collection.extract_collection_attributes(),
            by default None.
        collection_null_attributes : dict[AttributeName, CollectionAttribute], optional
            Optional memoization of collection.extract_null_attributes(),
            by default None.

        Returns
        -------
        float
            the collection entropy

        Limitations
        -----------
            Does not take into account non-String attributes during scoring.
        """
        attributes: dict[str, list[CollectionAttribute]] = (
            collection_attributes or collection.extract_collection_attributes()
        )
        null_attributes: dict[str, CollectionAttribute] = (
            collection_null_attributes or collection.extract_null_attributes()
        )

        # Create a list of all probabilities for every attribute name/value pair,
        # in the order of collection.attributes_frequency_counts.items()
        collection_probabilities: list[float] = []
        for attr_name, attr_values in attributes.items():
            if attr_name in null_attributes:
                null_attr = null_attributes[attr_name]
                attr_values.append(null_attr)

            # Create an array of the probability of all possible attr_name/value combos
            # existing in the collection
            collection_probabilities.extend(
                [
                    attr_value.total_tokens / collection.token_total_supply
                    for attr_value in attr_values
                ]
            )

        logger.debug("Calculated collection probabilties: %s", collection_probabilities)
        collection_entropy = -np.dot(
            collection_probabilities, np.log2(collection_probabilities)
        )

        return collection_entropy
